#!/usr/bin/env python

"""
Data classes.

Copyright (C) 2007, 2008, 2009 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.

--------

The central data structure classes in this module are the following:

  * Class
  * Function
  * Module

All of the above support the Naming interface either explicitly or through
general conformance, meaning that all can be asked to provide their 'full_name'
using the method of that name.

Additionally, all of the above also support a dictionary interface in order to
access names within their defined scopes. Specific methods also exist in order
to distinguish between certain kinds of attributes:

  * Class: class_attributes, all_class_attributes, instance_attributes, all_attributes
  * Function: parameters, locals, all_locals
  * Module: module_attributes

These specific methods are useful in certain situations.

The above classes also provide an 'astnode' attribute, indicating the AST node
where each such object is defined.
"""

from micropython.program import DataObject, ReplaceableContext, PlaceholderContext

def shortrepr(obj):
    if obj is None:
        return repr(None)
    else:
        return obj.__shortrepr__()

# Special representations.

class AtLeast:

    "A special representation for numbers of a given value or greater."

    def __init__(self, count):
        self.count = count

    def __eq__(self, other):
        return 0

    __lt__ = __le__ = __eq__

    def __ne__(self, other):
        return 1

    def __gt__(self, other):
        if isinstance(other, AtLeast):
            return 0
        else:
            return self.count > other

    def __ge__(self, other):
        if isinstance(other, AtLeast):
            return 0
        else:
            return self.count >= other

    def __iadd__(self, other):
        if isinstance(other, AtLeast):
            self.count += other.count
        else:
            self.count += other
        return self

    def __radd__(self, other):
        if isinstance(other, AtLeast):
            return AtLeast(self.count + other.count)
        else:
            return AtLeast(self.count + other)

    def __repr__(self):
        return "AtLeast(%r)" % self.count

# Mix-ins and abstract classes.

class Naming:

    "A mix-in providing naming conveniences."

    def full_name(self):
        if self.name is not None:
            return self.parent.full_name() + "." + self.name
        else:
            return self.parent.full_name()

class NamespaceDict:

    "A mix-in providing dictionary methods."

    def __init__(self, module=None):
        self.namespace = {}
        self.globals = set()
        self.module = module
        self.finalised = 0

    def __delitem__(self, name):
        del self.namespace[name]

    def has_key(self, name):
        return self.namespace.has_key(name)

    def keys(self):
        return self.namespace.keys()

    def values(self):
        return self.namespace.values()

    def items(self):
        return self.namespace.items()

    def __getitem__(self, name):
        return self.namespace[name]

    def get(self, name, default=None):
        return self.namespace.get(name, default)

    def __setitem__(self, name, value):
        self.set(name, value)

    def set(self, name, value, single_assignment=1):

        """
        A more powerful set operation, making 'name' refer to 'value' whilst
        indicating whether a 'single_assignment' (true by default) occurs in
        this operation (or whether the operation covers potentially many
        assignments in the lifetime of a program).
        """

        if name in self.globals:
            self.module.set(name, value, 0)
        else:
            self._set(name, value, single_assignment)

    def set_module(self, name, value):

        """
        A specialised set operation, making 'name' refer to 'value' in the
        context of making a module reference available in association with
        'name' as part of the import of that module or a submodule of that
        module.
        """

        self._set(name, value, 1)

    def _set(self, name, attr_or_value, single_assignment=1):

        """
        The underlying set operation associating 'name' with the given
        'attr_or_value'.
        See: docs/assignment.txt
        """

        # Add and/or obtain the namespace entry.

        if not self.namespace.has_key(name):
            self.namespace[name] = Attr(None, self, name)

        attr = self.namespace[name]

        # Handle attribute assignment as well as assignment of basic objects.
        # Attempt to fix the context if not explicitly defined.

        if isinstance(attr_or_value, Attr):
            context_values = self.get_updated_context_values(attr_or_value.context_values)
        else:
            context_values = self.get_updated_context_values([self.get_context_and_value(attr_or_value)])

        attr.update(context_values, single_assignment)

    def get_context_and_value(self, value):

        "Return a context, value tuple for the given 'value'."

        # Functions have a replaceable context.

        if isinstance(value, Function):
            return (ReplaceableContext, value)

        # Classes use placeholder contexts which cannot be replaced but which
        # do not communicate useful contextual information.

        elif isinstance(value, Class):
            return (PlaceholderContext, value)

        # Other values employ themselves as the context.

        else:
            return (value, value)

    def get_updated_context_values(self, context_values):

        """
        Adapt the contexts found in the given 'context_values', returning a new
        set.
        See: docs/assignment.txt
        """

        results = set()

        for context, value in context_values:

            # Set the context of instances to themselves.

            if isinstance(value, Instance):
                results.add((value, value))
            else:
                results.add((context, value))

        return results

    def make_global(self, name):
        if not self.namespace.has_key(name):
            self.globals.add(name)
            return 1
        else:
            return 0

    def attributes_as_list(self):

        "Return the attributes in a list."

        self.finalise_attributes()
        l = [None] * len(self.keys())
        for attr in self.values():
            l[attr.position] = attr
        return l

    def finalise_attributes(self):

        "Make sure all attributes are fully defined."

        if self.finalised:
            return

        # The default action is to assign attribute positions sequentially.

        for i, attr in enumerate(self.values()):
            attr.position = i

        self.finalised = 1

# Program data structures. There are two separate kinds of structures: those
# with context, which are the values manipulated by programs, and those without
# context, which are typically constant things which are stored alongside the
# program but which are wrapped in context-dependent structures in the running
# program.

class Attr:

    "An attribute entry having a context."

    def __init__(self, position, parent, name):

        """
        Initialise the attribute with the given 'position' within the collection
        of attributes of its 'parent', indicating its 'name'.
        """

        self.position = position
        self.parent = parent
        self.name = name

        self.context_values = set()

        # Number of assignments per name.

        self.assignments = None
        self.referenced = 0

    def set_referenced(self):

        "Indicate that the contents are referenced via a namespace."

        self.referenced = 1

        for value in self.get_values():
            if value is not None:
                value.set_referenced()

    def get_contexts(self):
        return [c for (c, v) in self.context_values]

    def get_values(self):
        return [v for (c, v) in self.context_values]

    def get_context(self):
        if len(self.context_values) == 1:
            return self.get_contexts()[0]
        else:
            return None

    def get_value(self):
        if len(self.context_values) == 1:
            return self.get_values()[0]
        else:
            return None

    def update(self, context_values, single_assignment):

        """
        Update the attribute, adding the 'context_values' provided to the
        known details associated with the attribute, changing the number of
        assignments according to the 'single_assignment' status of the
        operation, where a true value indicates that only one assignment is
        associated with the update, and a false value indicates that potentially
        many assignments may be involved.
        """

        if self.assignments is None:
            if single_assignment:
                self.assignments = 1
            else:
                self.assignments = AtLeast(1)
        else:
            if single_assignment:
                self.assignments += 1
            else:
                self.assignments += AtLeast(1)

        # Communicate referencing information.

        if self.referenced:
            for context, value in context_values:
                if value is not None:
                    value.set_referenced()

        self.context_values.update(context_values)

    def is_class_attribute(self):
        return isinstance(self.parent, Class)

    def defined_within_hierarchy(self):

        """
        Return whether the parent and context of the attribute belong to the
        same class hierarchy.
        """

        # Must be defined within a class.

        if isinstance(self.parent, Class):

            # To be sure, all contexts must be classes and be the same as the
            # parent, or be a superclass of the parent, or be a subclass of the
            # parent.

            for context in self.get_contexts():
                if not (
                    isinstance(context, Class) and (
                    context is self.parent or
                    context.has_subclass(self.parent) or
                    self.parent.has_subclass(context))
                    ):
                    return 0

            return 1

        # Instance attributes are not defined within a hierarchy.

        else:
            return 0

    def defined_outside_hierarchy(self):

        """
        Return whether the parent and context of the attribute never belong to
        the same class hierarchy.
        """

        # Must be defined within a class.

        if isinstance(self.parent, Class):

            # To be sure, all contexts must be classes and be the same as the
            # parent, or be a superclass of the parent, or be a subclass of the
            # parent.

            for context in self.get_contexts():
                if not (
                    isinstance(context, Class) and not (
                    context is self.parent or
                    context.has_subclass(self.parent) or
                    self.parent.has_subclass(context))
                    ):
                    return 0

            return 1

        # Instance attributes are not defined within a hierarchy.

        else:
            return 0

    def __repr__(self):
        return "Attr(%r, %s, %r) # [%s], %r" % (
            self.position, shortrepr(self.parent), self.name,
            self._context_values_str(), self.assignments
            )

    def _context_values_str(self):
        l = []
        for (c, v) in self.context_values:
            l.append("(c=%s, v=%s)" % (shortrepr(c), shortrepr(v)))
        return ", ".join(l)

    def as_raw(self, objtable, paramtable):
        return [
            (
                self.get_context() and self.get_context().location,
                self.get_value() and self.get_value().location
                )
            ]

# Instances are special in that they need to be wrapped together with context in
# a running program, but they are not generally constant.

class Instance:

    "A placeholder indicating the involvement of an instance."

    def __init__(self):
        self.parent = None
        self.referenced = 0

        # Image generation details.

        self.location = None

    def set_referenced(self):
        self.referenced = 1

    def __repr__(self):
        return "Instance()"

    __shortrepr__ = __repr__

class Constant:

    "A superclass for all constant or context-free structures."

    pass

# Data objects appearing in programs before run-time.

class Const(Constant, Instance):

    "A constant object with no context."

    def __init__(self, value):
        Instance.__init__(self)
        self.value = value

    def get_value(self):
        return value

    def __repr__(self):
        if self.location is not None:
            return "Const(%r, location=%r)" % (self.value, self.location)
        else:
            return "Const(%r)" % self.value

    __shortrepr__ = __repr__

    def as_raw(self, objtable, paramtable):
        # NOTE: Need class details!
        return [
            DataObject(
                objtable.as_list().get_code(self.value_type_name()),
                objtable.get_index(self.value_type_name()), # is instance
                None,
                self.value_type_name(),
                1 # size
                )
            ] + self.raw_data()

    def raw_data(self):
        # NOTE: Start simple and use single entries for most types.
        if self.value_type_name() in ("__builtins__.tuple", "__builtins__.list"):
            return [len(self.value)] + list(self.value)
        else:
            return [self.value]

    # Support constants as dictionary keys in order to build constant tables.

    def __eq__(self, other):
        return other is not None and self.value == other.value and self.value.__class__ is other.value.__class__

    def __hash__(self):
        return hash(self.value)

    def value_type_name(self):
        return "__builtins__." + self.value.__class__.__name__

class Class(NamespaceDict, Naming, Constant):

    "An inspected class."

    def __init__(self, name, parent, module=None, node=None):

        """
        Initialise the class with the given 'name', 'parent' object, optional
        'module' and optional AST 'node'.
        """

        NamespaceDict.__init__(self, module)
        self.name = name
        self.parent = parent
        self.astnode = node
        self.referenced = 0

        # Superclasses, descendants and attributes.

        self.bases = []
        self.descendants = set()
        self.instattr = set()           # instance attributes
        self.relocated = set()          # attributes which do not have the same
                                        # position as those of the same name in
                                        # some superclasses

        # Caches.

        self.all_instattr = None        # cache for instance_attributes
        self.all_instattr_names = None  # from all_instattr
        self.all_classattr = None       # cache for all_class_attributes
        self.all_classattr_names = None # from all_classattr
        self.allattr = None             # cache for all_attributes
        self.allattr_names = None       # from allattr

        # Image generation details.

        self.location = None
        self.code_location = None
        self.instantiator = None
        self.instance_template_location = None # for creating instances at run-time

        # Program-related details.

        self.blocks = None
        self.temp_usage = 0
        self.local_usage = 0
        self.all_local_usage = 0

        # Add this class to its attributes.

        self.set("__class__", self)

    def set_referenced(self):
        self.referenced = 1

    def __repr__(self):
        if self.location is not None:
            return "Class(%r, %s, location=%r)" % (self.name, shortrepr(self.parent), self.location)
        else:
            return "Class(%r, %s)" % (self.name, shortrepr(self.parent))

    def __shortrepr__(self):
        return "Class(%r, %s)" % (self.name, shortrepr(self.parent))

    def as_raw(self, objtable, paramtable, with_instantiator=1):
        classcode = objtable.as_list().get_code(self.full_name())
        attrcode = objtable.get_index(self.full_name())

        # Include a template of an instance for use when instantiating classes.

        call_method = self.all_class_attributes().get("__call__")
        call_method_value = call_method and call_method.get_value()
        call_method_code_location = call_method_value and call_method_value.code_location
        call_method_funccode = call_method_value and paramtable.as_list().get_code(call_method_value.full_name())

        instantiator_funccode = paramtable.as_list().get_code(self.get_instantiator().full_name())

        # NOTE: The instantiator code is the first block of the class.

        if not with_instantiator:
            instantiator_code_location = self.full_name()
        else:
            instantiator_code_location = self.get_instantiator().blocks[0].location

        return [

            # Template instance...

            DataObject(
                classcode,
                attrcode, # is instance
                call_method_code_location,
                self.full_name(),
                len(self.instance_attributes()) + 1, # size
                call_method_funccode # funccode
                ),

            # Class...

            DataObject(
                classcode,
                None, # is not instance
                instantiator_code_location,
                self.full_name(),
                len(self.class_attributes()) + 1, # size
                instantiator_funccode # funccode
                )
            ]

    def get_body_block(self):
        return self.get_instantiator().blocks[0]

    # Namespace-related methods.

    def get_updated_context_values(self, context_values):

        """
        Adapt the contexts found in the given 'context_values', returning a new
        set.
        See: docs/assignment.txt
        """

        results = set()

        for context, value in context_values:

            # Change the ownership of functions.

            if context is ReplaceableContext and value is not None and isinstance(value, Function):
                results.add((self, value))
            else:
                results.add((context, value))

        return NamespaceDict.get_updated_context_values(self, results)

    def finalise_attributes(self):

        "Make sure that all attributes are fully defined."

        if self.finalised:
            return

        self.finalise_class_attributes()
        self.finalise_instance_attributes()
        self.finalised = 1

    # Convenience methods for accessing functions and methods.

    def get_instantiator(self):

        "Return a function which can be used to instantiate the class."

        if self.instantiator is None:
            self.instantiator = self.get_init_method().as_instantiator()
        return self.instantiator

    def get_init_method(self):
        return self.all_class_attributes()["__init__"].get_value()

    # Class-specific methods.

    def add_base(self, base):
        self.bases.append(base)
        base.add_descendant(self)

    def add_instance_attribute(self, name):
        self.instattr.add(name)

    def add_descendant(self, cls):
        self.descendants.add(cls)
        for base in self.bases:
            base.add_descendant(cls)

    def has_subclass(self, other):
        return other in self.descendants

    def all_descendants(self):
        d = {}
        for cls in self.descendants:
            d[cls.full_name()] = cls
        return d

    "Return the attribute names provided by this class only."

    class_attribute_names = NamespaceDict.keys

    def class_attributes(self):

        "Return class attributes provided by this class only."

        return dict(self)

    def all_class_attribute_names(self):

        "Return the attribute names provided by classes in this hierarchy."

        if self.all_classattr_names is None:
            self.all_class_attributes()
        return self.all_classattr_names

    def all_class_attributes(self):

        "Return all class attributes, indicating the class which provides them."

        self.finalise_class_attributes()
        return self.all_classattr

    def finalise_class_attributes(self):

        "Make sure that the class attributes are fully defined."

        if self.all_classattr is None:
            self.all_classattr = {}
            clsattr = {}

            # Record provisional position information for attributes of this
            # class.

            for name in self.class_attributes().keys():
                clsattr[name] = set() # position not yet defined

            reversed_bases = self.bases[:]
            reversed_bases.reverse()

            # For the bases in reverse order, acquire class attribute details.

            for cls in reversed_bases:
                for name, attr in cls.all_class_attributes().items():
                    self.all_classattr[name] = attr

                    # Record previous attribute information.

                    if clsattr.has_key(name):
                        clsattr[name].add(attr.position)

            # Record class attributes provided by this class and its bases,
            # along with their positions.

            self.all_classattr.update(self.class_attributes())

            if clsattr:
                for i, name in enumerate(self._get_position_list(clsattr)):
                    self.all_classattr[name].position = i

        return self.all_classattr

    def instance_attribute_names(self):

        "Return the instance attribute names provided by the class."

        if self.all_instattr_names is None:
            self.instance_attributes()
        return self.all_instattr_names

    def instance_attributes(self):

        "Return instance-only attributes for instances of this class."

        self.finalise_instance_attributes()
        return self.all_instattr

    def finalise_instance_attributes(self):

        "Make sure that the instance attributes are fully defined."

        # Cache the attributes by converting the positioned attributes into a
        # dictionary.

        if self.all_instattr is None:
            self.all_instattr = self._get_attributes()
            self.all_instattr_names = self.all_instattr.keys()

        return self.all_instattr

    def _get_attributes(self):

        """
        Return a dictionary mapping names to Attr instances incorporating
        information about their positions in the final instance structure.
        """

        instattr = {}

        # Record provisional position information for attributes of this
        # instance.

        for name in self.instattr:
            instattr[name] = set() # position not yet defined

        reversed_bases = self.bases[:]
        reversed_bases.reverse()

        # For the bases in reverse order, acquire instance attribute
        # details.

        for cls in reversed_bases:
            for name, attr in cls.instance_attributes().items():

                # Record previous attribute information.

                if instattr.has_key(name):
                    instattr[name].add(attr.position)
                else:
                    instattr[name] = set([attr.position])

        # Build the dictionary of attributes using the existing positions known
        # for each name.

        d = {}
        for i, name in enumerate(self._get_position_list(instattr)):
            d[name] = Attr(i, Instance(), name)
        return d

    def _get_position_list(self, positions):

        """
        Return a list of attribute names for the given 'positions' mapping from
        names to positions, indicating the positions of the attributes in the
        final instance structure.
        """

        position_items = positions.items()
        namearray = [None] * len(position_items)

        # Get the positions in ascending order of list size, with lists
        # of the same size ordered according to their smallest position
        # value.

        position_items.sort(self._cmp_positions)

        # Get the names in position order.

        held = []

        for name, pos in position_items:
            pos = list(pos)
            pos.sort()
            if pos and pos[0] < len(namearray) and namearray[pos[0]] is None:
                namearray[pos[0]] = name
            else:
                if pos:
                    self.relocated.add(name)
                held.append((name, pos))

        for i, attr in enumerate(namearray):
            if attr is None:
                name, pos = held.pop()
                namearray[i] = name

        #print self.name, positions
        #print "->", namearray
        return namearray

    def _cmp_positions(self, a, b):

        "Compare name plus position list operands 'a' and 'b'."

        name_a, list_a = a
        name_b, list_b = b
        if len(list_a) < len(list_b):
            return -1
        elif len(list_a) > len(list_b):
            return 1
        elif not list_a:
            return 0
        else:
            return cmp(min(list_a), min(list_b))

    def all_attribute_names(self):

        """
        Return the names of all attributes provided by instances of this class.
        """

        self.allattr_names = self.allattr_names or self.all_attributes().keys()
        return self.allattr_names

    def all_attributes(self):

        """
        Return all attributes for an instance, indicating either the class which
        provides them or that the instance itself provides them.
        """

        if self.allattr is None:
            self.allattr = {}
            self.allattr.update(self.all_class_attributes())
            for name, attr in self.instance_attributes().items():
                if self.allattr.has_key(name):
                    print "Instance attribute %r in %r overrides class attribute." % (name, self)
                self.allattr[name] = attr
        return self.allattr

class Function(NamespaceDict, Naming, Constant):

    "An inspected function."

    def __init__(self, name, parent, argnames, defaults, has_star, has_dstar, module=None, node=None):

        """
        Initialise the function with the given 'name', 'parent', list of
        'argnames', list of 'defaults', the 'has_star' flag (indicating the
        presence of a * parameter), the 'has_dstar' flag (indicating the
        presence of a ** parameter), optional 'module', and optional AST 'node'.
        """

        NamespaceDict.__init__(self, module)
        self.name = name
        self.parent = parent
        self.argnames = argnames
        self.defaults = defaults
        self.has_star = has_star
        self.has_dstar = has_dstar
        self.astnode = node
        self.referenced = 0

        # For lambda functions with defaults, add a context argument.

        if name is None and defaults:
            self.argnames.insert(0, "<context>")

        # Initialise the positional names.

        self.positional_names = self.argnames[:]
        if has_dstar:
            self.dstar_name = self.positional_names[-1]
            del self.positional_names[-1]
        if has_star:
            self.star_name = self.positional_names[-1]
            del self.positional_names[-1]

        # Initialise default storage.
        # NOTE: This must be initialised separately due to the reliance on node
        # NOTE: visiting.

        self.default_attrs = []

        # Caches.

        self.localnames = None # cache for locals

        # Add parameters to the namespace.

        self._add_parameters(argnames)

        # Image generation details.

        self.location = None
        self.code_location = None

        # Program-related details.

        self.blocks = None
        self.body_block = None

        self.temp_usage = 0
        self.local_usage = 0
        self.all_local_usage = 0

    def set_referenced(self):
        self.referenced = 1

    def _add_parameters(self, argnames):
        for name in argnames:
            if isinstance(name, tuple):
                self._add_parameters(name)
            else:
                self.set(name, None)

    def __repr__(self):
        if self.location is not None:
            return "Function(%r, %s, %r, location=%r, code_location=%r)" % (
                self.name, shortrepr(self.parent), self.argnames, self.location, self.code_location
                )
        else:
            return "Function(%r, %s, %r)" % (
                self.name, shortrepr(self.parent), self.argnames
                )

    def __shortrepr__(self):
        return "Function(%r, %s)" % (
            self.name, shortrepr(self.parent)
            )

    def as_raw(self, objtable, paramtable):
        # NOTE: Need class and parameter details! Should arguably be an instance of types.FunctionType.
        return [
            DataObject(
                objtable.as_list().get_code("__builtins__.function"),
                objtable.get_index("__builtins__.function"), # is instance
                self.code_location,
                "__builtins__.function",
                len(self.defaults) + 1, # size (not accurate for lambda functions before instantiation)
                paramtable.as_list().get_code(self.full_name()) # funccode
                )
            ]

    def get_body_block(self):
        return self.body_block

    # Namespace-related methods.

    def store_default(self, value):
        attr = Attr(None, self, None)
        attr.update([self.get_context_and_value(value)], 1)
        self.default_attrs.append(attr)

    def make_global(self, name):
        if name not in self.argnames and not self.has_key(name):
            self.globals.add(name)
            return 1
        else:
            return 0

    def parameters(self):

        """
        Return a dictionary mapping parameter names to their position in the
        parameter list.
        """

        parameters = {}
        for i, name in enumerate(self.argnames):
            parameters[name] = i
        return parameters

    def all_locals(self):

        "Return a dictionary mapping names to local and parameter details."

        return dict(self)

    def locals(self):

        "Return a dictionary mapping names to local details."

        if self.localnames is None:
            self.localnames = {}
            self.localnames.update(self.all_locals())
            for name in self.argnames:
                del self.localnames[name]
        return self.localnames

    def is_method(self):

        "Return whether this function is a method."

        return isinstance(self.parent, Class)

    def is_relocated(self, name):

        """
        Determine whether the given attribute 'name' is relocated for instances
        having this function as a method.
        """

        for cls in self.parent.descendants:
            if name in cls.relocated:
                return 1
        return 0

    def finalise_attributes(self):

        """
        Make sure all attributes (locals) are fully defined. Note that locals
        are not attributes in the sense of class, module or instance attributes.
        Defaults are also finalised by this method.
        """

        if self.finalised:
            return

        # Defaults.

        for i, default in enumerate(self.default_attrs):
            default.position = i

        # Locals.

        i = None
        for i, name in enumerate(self.argnames):
            self[name].position = i

        if i is not None:
            nparams = i + 1
        else:
            nparams = 0

        i = None
        for i, attr in enumerate(self.locals().values()):
            attr.position = i + nparams

        if i is not None:
            nothers = i + 1
        else:
            nothers = 0

        self.local_usage = nothers
        self.all_local_usage = nparams + nothers
        self.finalised = 1

    def as_instantiator(self):

        "Make an instantiator function from a method, keeping all arguments."

        function = Function(self.parent.name, self.parent.parent, self.argnames, self.defaults,
            self.has_star, self.has_dstar, self.module, self.astnode)
        function.default_attrs = self.default_attrs
        return function

class UnresolvedName(NamespaceDict, Constant):

    "A module, class or function which was mentioned but could not be imported."

    def __init__(self, name, parent_name, module=None):
        NamespaceDict.__init__(self, module)
        self.name = name
        self.parent_name = parent_name
        self.parent = None
        self.referenced = 0

        self.descendants = set()

    def set_referenced(self):
        self.referenced = 1

    def add_descendant(self, cls):
        self.descendants.add(cls)

    def all_class_attributes(self):
        return {}

    def instance_attributes(self):
        return {}

    def __repr__(self):
        return "UnresolvedName(%r, %r)" % (self.name, self.parent_name)

    __shortrepr__ = __repr__

    def full_name(self):
        if self.name is not None:
            return self.parent_name + "." + self.name
        else:
            return self.parent_name

class Module(NamespaceDict, Constant):

    "An inspected module's core details."

    def __init__(self, name):
        NamespaceDict.__init__(self, self)
        self.name = name
        self.parent = None
        self.referenced = 0

        # Original location details.

        self.astnode = None

        # Complete lists of classes and functions.

        self.all_objects = set()

        # Keyword records.

        self.keyword_names = set()

        # Image generation details.

        self.location = None
        self.code_location = None

        # Program-related details.

        self.blocks = None
        self.temp_usage = 0
        self.local_usage = 0
        self.all_local_usage = 0

    def set_referenced(self):
        self.referenced = 1

    def full_name(self):
        return self.name

    def __repr__(self):
        if self.location is not None:
            return "Module(%r, location=%r)" % (self.name, self.location)
        else:
            return "Module(%r)" % self.name

    def __shortrepr__(self):
        return "Module(%r)" % self.name

    def as_raw(self, objtable, paramtable):
        return [
            DataObject(
                objtable.as_list().get_code(self.full_name()),
                None, # NOTE: module name not used as an attribute, but should be instance
                None,
                self.full_name(),
                len(self.module_attributes()) + 1 # size
                )
            ]

    # Attribute methods.

    "Return the module attribute names provided by the module."

    module_attribute_names = NamespaceDict.keys

    def module_attributes(self):

        "Return a dictionary mapping names to module attributes."

        return dict(self)

# vim: tabstop=4 expandtab shiftwidth=4
